@@ -588,7 +588,9 @@ fn load_subs_from_file(sbr: &Subrandr, path: &Path) -> Result<subrandr::Subtitle
     Ok(match path.extension().and_then(|x| x.to_str()) {
         Some("srv3" | "ytt") => {
             let document = subrandr::srv3::parse(sbr, &std::fs::read_to_string(path).unwrap())?;
-            Subtitles::Srv3(util::rc::Rc::new(subrandr::srv3::convert(sbr, document)))
+            Subtitles::Srv3(util::rc::Rc::new(subrandr::srv3::convert(
+                sbr, document, None,
+            )))
         }
         Some("vtt") => {
             let text = std::fs::read_to_string(path).unwrap();
