@@ -15,8 +15,8 @@ use crate::{
 
 mod blit;
 pub(super) mod blur;
+use blur::*;
 mod scale;
-use blur::gaussian_sigma_to_box_radius;
 mod strip;
 pub use strip::*;
 
@@ -215,6 +215,27 @@ fn fill_rect<B: BlendMode<P>, P: DrawPixel>(
     }
 }
 
+#[track_caller]
+fn check_pixel_buffer_dimensions(
+    buffer_len: usize,
+    width: u32,
+    height: u32,
+    stride: u32,
+    name: &std::fmt::Arguments,
+) {
+    let Some(n_pixels) = (height as usize).checked_mul(stride as usize) else {
+        panic!("Size passed to {name} overflows `height * stride`",);
+    };
+    assert!(
+        buffer_len >= n_pixels,
+        "Buffer passed to {name} is too small",
+    );
+    assert!(
+        width <= stride,
+        "width passed to {name} is larger than stride",
+    );
+}
+
 pub struct RenderTargetView<'a, P> {
     buffer: &'a mut [P],
     width: u32,
@@ -224,17 +245,12 @@ pub struct RenderTargetView<'a, P> {
 
 impl<'a, P> RenderTargetView<'a, P> {
     pub fn new(buffer: &'a mut [P], width: u32, height: u32, stride: u32) -> Self {
-        let self_name = std::any::type_name::<Self>();
-        let Some(n_pixels) = (height as usize).checked_mul(stride as usize) else {
-            panic!("Size passed to {self_name}::new overflows `height * stride`",);
-        };
-        assert!(
-            buffer.len() >= n_pixels,
-            "Buffer passed to {self_name}::new is too small",
-        );
-        assert!(
-            width <= stride,
-            "width passed to {self_name}::new is larger than stride",
+        check_pixel_buffer_dimensions(
+            buffer.len(),
+            width,
+            height,
+            stride,
+            &format_args!("{}::new", std::any::type_name::<Self>()),
         );
 
         Self {
@@ -394,9 +410,9 @@ impl TexturePixel for u8 {
 
 #[derive(Clone, Hash, PartialEq, Eq)]
 pub struct Texture<'a> {
+    data: TextureData<'a>,
     width: u32,
     height: u32,
-    data: TextureData<'a>,
 }
 
 impl Texture<'static> {
@@ -440,7 +456,23 @@ impl Texture<'static> {
     };
 }
 
-impl Texture<'_> {
+impl<'a> Texture<'a> {
+    pub fn new_borrowed_mono(data: &'a [u8], width: u32, height: u32) -> Self {
+        check_pixel_buffer_dimensions(
+            data.len(),
+            width,
+            height,
+            width,
+            &format_args!("{}::new_borrowed_mono", std::any::type_name::<Self>()),
+        );
+
+        Self {
+            data: TextureData::BorrowedMono(data),
+            width,
+            height,
+        }
+    }
+
     pub fn size(&self) -> Vec2<u32> {
         Vec2::new(self.width, self.height)
     }
@@ -616,6 +648,36 @@ impl Rasterizer {
         }
     }
 
+    pub fn scale_mono_raw(
+        &self,
+        target: RenderTargetView<'_, MaybeUninit<u8>>,
+        src: &[u8],
+        src_width: u32,
+        src_height: u32,
+        src_stride: usize,
+        src_off: Vec2<i32>,
+        src_size: Vec2<i32>,
+    ) {
+        scale::scale_mono(
+            target, src, src_stride, src_width, src_height, src_off, src_size,
+        );
+    }
+
+    pub fn scale_bgra_raw(
+        &self,
+        target: RenderTargetView<'_, MaybeUninit<Premultiplied<BGRA8>>>,
+        src: &[Premultiplied<BGRA8>],
+        src_width: u32,
+        src_height: u32,
+        src_stride: usize,
+        src_off: Vec2<i32>,
+        src_size: Vec2<i32>,
+    ) {
+        scale::scale_bgra(
+            target, src, src_stride, src_width, src_height, src_off, src_size,
+        );
+    }
+
     pub fn scale_texture(
         &self,
         texture: &Texture,
@@ -689,11 +751,17 @@ impl super::Rasterizer for Rasterizer {
     fn blur_texture(&mut self, texture: &super::Texture, blur_sigma: f32) -> super::BlurOutput {
         let texture = unwrap_sw_texture(texture);
 
-        self.blurer.prepare(
-            texture.width as usize,
-            texture.height as usize,
-            gaussian_sigma_to_box_radius(blur_sigma),
-        );
+        let is_box_blur;
+        let kernel = if blur_sigma > 2.0 {
+            is_box_blur = true;
+            BlurKernel::Box(BoxBlurKernel::from_gaussian_stddev(blur_sigma))
+        } else {
+            is_box_blur = false;
+            BlurKernel::Gaussian(GaussianBlurKernel::new(blur_sigma))
+        };
+
+        self.blurer
+            .prepare(texture.width as usize, texture.height as usize, kernel);
 
         let dx = self.blurer.padding() as i32;
         let dy = self.blurer.padding() as i32;
@@ -726,12 +794,17 @@ impl super::Rasterizer for Rasterizer {
             ),
         }
 
-        self.blurer.box_blur_horizontal();
-        self.blurer.box_blur_horizontal();
-        self.blurer.box_blur_horizontal();
-        self.blurer.box_blur_vertical();
-        self.blurer.box_blur_vertical();
-        self.blurer.box_blur_vertical();
+        if is_box_blur {
+            self.blurer.blur_horizontal();
+            self.blurer.blur_horizontal();
+            self.blurer.blur_horizontal();
+            self.blurer.blur_vertical();
+            self.blurer.blur_vertical();
+            self.blurer.blur_vertical();
+        } else {
+            self.blurer.blur_horizontal();
+            self.blurer.blur_vertical();
+        }
 
         let result = Texture::new_with(Vec2::new(width as u32, height as u32), |target| {
             blit::copy_float_to_mono(target, self.blurer.front(), width, width, height, 0, 0);
