@@ -25,6 +25,7 @@ struct DebugFlags {
     draw_version_string: bool,
     draw_perf_info: bool,
     dpi_override: Option<u32>,
+    srv3_use_inlines: bool,
 }
 
 impl DebugFlags {
@@ -36,6 +37,7 @@ impl DebugFlags {
                 match token {
                     "draw_version" => result.draw_version_string = true,
                     "draw_perf" => result.draw_perf_info = true,
+                    "srv3_use_inlines" => result.srv3_use_inlines = true,
                     #[allow(clippy::single_match)]
                     _ => match token.split_once("=") {
                         Some(("override_dpi", value_str)) => {
