@@ -21,6 +21,7 @@ use crate::{
     display::DisplayPass,
     layout::{
         self,
+        block::BlockContainerFragment,
         inline::{InlineContentBuilder, InlineContentFragment},
         FixedL, LayoutConstraints, LayoutContext, Point2L,
     },
@@ -45,10 +46,6 @@ pub struct SubtitleContext {
 }
 
 impl SubtitleContext {
-    pub fn ppi(&self) -> u32 {
-        self.dpi * 96 / 72
-    }
-
     pub fn pixel_scale(&self) -> f32 {
         self.dpi as f32 / 72.0
     }
@@ -86,7 +83,11 @@ pub(crate) struct FrameLayoutPass<'s, 'frame> {
     pub lctx: &'frame mut LayoutContext<'frame, 's>,
     pub t: u32,
     unchanged_range: Range<u32>,
-    fragments: Vec<(Point2L, InlineContentFragment)>,
+    fragments: Vec<(Point2L, BlockContainerFragment)>,
+    // TODO: proper configuration system
+    // TODO: fully inline mode could probably keep using blocks for ruby which would clean
+    //       up a sizing hack in inline layout
+    pub srv3_use_inlines: bool,
 }
 
 impl FrameLayoutPass<'_, '_> {
@@ -116,7 +117,7 @@ impl FrameLayoutPass<'_, '_> {
         }
     }
 
-    pub fn emit_fragment(&mut self, pos: Point2L, block: InlineContentFragment) {
+    pub fn emit_fragment(&mut self, pos: Point2L, block: BlockContainerFragment) {
         self.fragments.push((pos, block));
     }
 }
@@ -450,23 +451,22 @@ impl Renderer<'_> {
         };
 
         if sbr.debug.draw_version_string {
-            let mut builder = InlineContentBuilder::new();
+            let mut builder = InlineContentBuilder::new(base_style.clone());
             let mut root = builder.root();
-            let mut main = root.push_span(base_style.clone());
 
-            main.push_text(concat!(
+            root.push_text(concat!(
                 "subrandr ",
                 env!("CARGO_PKG_VERSION"),
                 env!("BUILD_REV_SUFFIX"),
                 "\n"
             ));
 
-            _ = writeln!(main, "subtitle class: {subtitle_class_name}");
+            _ = writeln!(root, "subtitle class: {subtitle_class_name}");
 
-            _ = writeln!(main, "=== rasterizer: {} ===", rasterizer.name());
-            _ = rasterizer.write_debug_info(&mut main);
-            if !main.current_run_text().ends_with('\n') {
-                main.push_text("\n");
+            _ = writeln!(root, "=== rasterizer: {} ===", rasterizer.name());
+            _ = rasterizer.write_debug_info(&mut root);
+            if !root.current_run_text().ends_with('\n') {
+                root.push_text("\n");
             }
 
             {
@@ -474,17 +474,16 @@ impl Renderer<'_> {
                 let (footprint_divisor, footprint_suffix) =
                     util::human_size_suffix(stats.total_memory_footprint);
 
-                _ = writeln!(main, "=== glyph cache stats ===");
+                _ = writeln!(root, "=== glyph cache stats ===");
                 _ = writeln!(
-                    main,
+                    root,
                     "approximate memory footprint: {:.3}{footprint_suffix}B",
                     stats.total_memory_footprint as f32 / footprint_divisor as f32
                 );
-                _ = writeln!(main, "total entries: {}", stats.total_entries);
-                _ = writeln!(main, "current generation: {}", stats.generation);
+                _ = writeln!(root, "total entries: {}", stats.total_entries);
+                _ = writeln!(root, "current generation: {}", stats.generation);
             }
 
-            drop(main);
             drop(root);
             fragments.push((
                 Point2L::ZERO,
@@ -492,31 +491,33 @@ impl Renderer<'_> {
                     lctx,
                     &LayoutConstraints::NONE,
                     &builder.finish(),
-                    HorizontalAlignment::Left,
                 )?),
             ));
         }
 
         if sbr.debug.draw_perf_info {
-            let mut builder = InlineContentBuilder::new();
+            let mut builder = InlineContentBuilder::new({
+                let mut style = base_style.clone();
+                *style.make_text_align_mut() = HorizontalAlignment::Right;
+                style
+            });
             let mut root = builder.root();
-            let mut main = root.push_span(base_style.clone());
 
             _ = writeln!(
-                main,
+                root,
                 "{:.2}x{:.2} dpi:{}",
                 ctx.video_width, ctx.video_height, ctx.dpi
             );
 
             _ = writeln!(
-                main,
+                root,
                 "l:{:.2} r:{:.2} t:{:.2} b:{:.2}",
                 ctx.padding_left, ctx.padding_right, ctx.padding_top, ctx.padding_bottom
             );
 
             if !perf.is_empty() {
                 let mut draw_times = |name: &str, times: &PerfTimes, color: BGRA8| {
-                    main.push_span({
+                    root.push_span({
                         let mut result = base_style.create_derived();
                         *result.make_color_mut() = color;
                         *result.make_font_weight_mut() = I16Dot16::new(700);
@@ -528,7 +529,7 @@ impl Renderer<'_> {
                     let last = times.last().unwrap();
                     let avg = times.avg_frame_time();
                     _ = writeln!(
-                        main,
+                        root,
                         " last={:.1}ms avg={:.1}ms ({:.1}/s) max={:.1}ms ({:.1}/s)",
                         last,
                         avg,
@@ -544,14 +545,9 @@ impl Renderer<'_> {
                 draw_times("display", &perf.display, BGRA8::LIME);
                 draw_times("raster", &perf.raster, BGRA8::ORANGERED);
 
-                drop(main);
                 drop(root);
-                let perf_text_fragment = layout::inline::layout(
-                    lctx,
-                    &LayoutConstraints::NONE,
-                    &builder.finish(),
-                    HorizontalAlignment::Right,
-                )?;
+                let perf_text_fragment =
+                    layout::inline::layout(lctx, &LayoutConstraints::NONE, &builder.finish())?;
                 let graph_y = perf_text_fragment.fbox.size_for_layout().y;
                 fragments.push((
                     Point2L::new(
@@ -624,6 +620,7 @@ impl Renderer<'_> {
                 t,
                 unchanged_range: 0..u32::MAX,
                 fragments: Vec::new(),
+                srv3_use_inlines: self.sbr.debug.srv3_use_inlines,
             };
 
             fragments = {
@@ -652,12 +649,10 @@ impl Renderer<'_> {
         }
 
         {
-            let mut pass = DisplayPass {
-                output: &mut self.scene,
-            };
+            let mut pass = DisplayPass::new(&mut self.scene, ctx.dpi);
 
             for &(pos, ref fragment) in &fragments {
-                pass.display_inline_content_fragment(pos, fragment);
+                pass.display_block_container_fragment(pos, fragment);
             }
 
             for &(pos, ref fragment) in &debug_overlay_fragments {
