@@ -1,23 +1,26 @@
 use std::{collections::HashMap, ops::Range};
 
+use icu_locale::{LanguageIdentifier, LocaleDirectionality};
 use rasterize::color::BGRA8;
 use util::{
-    math::{I16Dot16, I26Dot6, Vec2f},
+    math::{I16Dot16, I26Dot6, Vec2},
     rc::Rc,
 };
 
 use crate::{
     layout::{
-        self, inline::InlineContentBuilder, FixedL, InlineLayoutError, LayoutConstraints, Point2L,
-        Vec2L,
+        self,
+        block::{BlockContainer, BlockContainerContent},
+        inline::{InlineContent, InlineContentBuilder, InlineSpanBuilder},
+        FixedL, InlineLayoutError, LayoutConstraints, Point2L, Vec2L,
     },
     log::{log_once_state, warning, LogOnceSet},
     renderer::FrameLayoutPass,
     srv3::{Event, RubyPosition},
     style::{
         computed::{
-            Alignment, FontSlant, HorizontalAlignment, InlineSizing, Length, Ruby, TextShadow,
-            VerticalAlignment,
+            Alignment, Direction, FontSlant, HorizontalAlignment, InlineSizing, Length, Ruby,
+            TextShadow, VerticalAlignment,
         },
         ComputedStyle,
     },
@@ -122,10 +125,10 @@ fn calculate_font_scale(
 
 fn font_scale_from_ctx(ctx: &SubtitleContext) -> f32 {
     calculate_font_scale(
-        ctx.pixels_to_css(ctx.video_width.into_f32()),
-        ctx.pixels_to_css(ctx.video_height.into_f32()),
-        ctx.pixels_to_css(ctx.player_width().into_f32()),
-        ctx.pixels_to_css(ctx.player_height().into_f32()),
+        Length::from_physical_pixels(ctx.video_width, ctx.dpi).to_f32(),
+        Length::from_physical_pixels(ctx.video_height, ctx.dpi).to_f32(),
+        Length::from_physical_pixels(ctx.player_width(), ctx.dpi).to_f32(),
+        Length::from_physical_pixels(ctx.player_height(), ctx.dpi).to_f32(),
     )
 }
 
@@ -139,94 +142,6 @@ fn font_size_to_pixels(size: u16) -> f32 {
     c
 }
 
-trait SubtitleContextCssExt {
-    // 1px = 1/96in
-    fn pixels_to_css(&self, physical_pixels: f32) -> f32;
-    fn pixels_from_css(&self, css_pixels: f32) -> f32;
-}
-
-impl SubtitleContextCssExt for SubtitleContext {
-    fn pixels_to_css(&self, physical_pixels: f32) -> f32 {
-        physical_pixels / self.pixel_scale()
-    }
-
-    fn pixels_from_css(&self, css_pixels: f32) -> f32 {
-        css_pixels * self.pixel_scale()
-    }
-}
-
-fn pixels_to_points(pixels: f32) -> f32 {
-    pixels * 96.0 / 72.0
-}
-
-#[derive(Debug, Clone)]
-pub struct Srv3TextShadow {
-    kind: EdgeType,
-    color: BGRA8,
-}
-
-impl Srv3TextShadow {
-    pub(crate) fn to_css(&self, ctx: &SubtitleContext, out: &mut Vec<TextShadow>) {
-        let a = font_scale_from_ctx(ctx) / 32.0;
-        let e = a.max(1.0);
-        let l = (2.0 * a).max(1.0);
-        let mut t = (3.0 * a).max(1.0);
-        let c = (5.0 * a).max(1.0);
-
-        match self.kind {
-            EdgeType::None => (),
-            EdgeType::HardShadow => {
-                // NOTE: If `text-shadow` is ever updated to use logical pixels instead of device
-                //       pixels this code should be updated.
-                //       Currently this increases the step on `dpi >= 144` but in that case
-                //       it would have to do the opposite and increase it if `dpi < 144`.
-                let step = (ctx.dpi >= 144) as i32 as f32 * 0.5 + 0.5;
-                let mut x = e;
-                while x <= t {
-                    out.push(TextShadow {
-                        offset: Vec2f::new(ctx.pixels_from_css(x), ctx.pixels_from_css(x)),
-                        blur_radius: I26Dot6::ZERO,
-                        color: self.color,
-                    });
-                    x += step;
-                }
-            }
-            EdgeType::Bevel => {
-                let offset = Vec2f::new(ctx.pixels_from_css(e), ctx.pixels_from_css(e));
-                out.push(TextShadow {
-                    offset,
-                    blur_radius: I26Dot6::ZERO,
-                    color: self.color,
-                });
-                out.push(TextShadow {
-                    offset: -offset,
-                    blur_radius: I26Dot6::ZERO,
-                    color: self.color,
-                });
-            }
-            EdgeType::Glow => out.extend(std::iter::repeat_n(
-                TextShadow {
-                    offset: Vec2f::ZERO,
-                    blur_radius: I26Dot6::from_f32(ctx.pixels_from_css(l)),
-                    color: self.color,
-                },
-                5,
-            )),
-            EdgeType::SoftShadow => {
-                let offset = Vec2f::new(ctx.pixels_from_css(l), ctx.pixels_from_css(l));
-                while t <= c {
-                    out.push(TextShadow {
-                        offset,
-                        blur_radius: I26Dot6::from_f32(ctx.pixels_from_css(t)),
-                        color: self.color,
-                    });
-                    t += a;
-                }
-            }
-        }
-    }
-}
-
 impl super::Point {
     pub fn to_alignment(self) -> Alignment {
         match self {
@@ -270,7 +185,8 @@ struct Window {
     //       How does a timestamp on a window work?
     //       Currently this is just ignored until I figure out what to do with it.
     range: Range<u32>,
-    alignment: Alignment,
+    segment_style: ComputedStyle,
+    vertical_align: VerticalAlignment,
     lines: Vec<VisualLine>,
 }
 
@@ -282,20 +198,88 @@ struct VisualLine {
 
 #[derive(Debug, Clone)]
 struct Segment {
+    pen: Pen,
     base_style: ComputedStyle,
-    font_size: u16,
     time_offset: u32,
     text: std::rc::Rc<str>,
-    shadow: Srv3TextShadow,
     ruby: Ruby,
 }
 
 impl Segment {
-    fn compute_style(&self, sctx: &SubtitleContext) -> ComputedStyle {
+    fn compute_shadows(&self, ctx: &SubtitleContext, out: &mut Vec<TextShadow>) {
+        let scale = FixedL::from_f32(font_scale_from_ctx(ctx) / 32.0);
+        let l1 = Length::from_pixels((scale).max(FixedL::ONE));
+        let l2 = Length::from_pixels((scale * 2).max(FixedL::ONE));
+        let l3 = Length::from_pixels((scale * 3).max(FixedL::ONE));
+        let l5 = Length::from_pixels((scale * 5).max(FixedL::ONE));
+        let primary_color = BGRA8::from_argb32(self.pen.edge_color.map_or_else(
+            || 0x222222 | (self.pen.foreground_color << 24),
+            |c| c | 0xFF000000,
+        ));
+
+        match self.pen.edge_type {
+            EdgeType::None => (),
+            EdgeType::HardShadow => {
+                let step = Length::HALF * (ctx.dpi < 144) as i32 + Length::HALF;
+                let mut x = l1;
+                while x <= l3 {
+                    out.push(TextShadow {
+                        offset: Vec2::new(x, x),
+                        blur_radius: Length::ZERO,
+                        color: primary_color,
+                    });
+                    x += step;
+                }
+            }
+            EdgeType::Bevel => {
+                // If there is no explicit edge color set then `Bevel` will use two
+                // distinct colors for the positive-offset and negative-offset shadow.
+                // The "inner" shadow will end up with a very light gray but the "outer" one
+                // will be the usual black.
+                let secondary_color = if self.pen.edge_color.is_none() {
+                    BGRA8::from_argb32(0xCCCCCC | (self.pen.foreground_color << 24))
+                } else {
+                    primary_color
+                };
+                let offset = Vec2::new(l1, l1);
+                out.push(TextShadow {
+                    offset,
+                    blur_radius: Length::ZERO,
+                    color: secondary_color,
+                });
+                out.push(TextShadow {
+                    offset: -offset,
+                    blur_radius: Length::ZERO,
+                    color: primary_color,
+                });
+            }
+            EdgeType::Glow => out.extend(std::iter::repeat_n(
+                TextShadow {
+                    offset: Vec2::ZERO,
+                    blur_radius: l2,
+                    color: primary_color,
+                },
+                5,
+            )),
+            EdgeType::SoftShadow => {
+                let offset = Vec2::new(l2, l2);
+                let mut x = l3;
+                while x <= l5 {
+                    out.push(TextShadow {
+                        offset,
+                        blur_radius: x,
+                        color: primary_color,
+                    });
+                    x += Length::from_pixels(scale);
+                }
+            }
+        }
+    }
+
+    fn compute_style(&self, sctx: &SubtitleContext, use_inlines: bool) -> ComputedStyle {
         let mut result = self.base_style.clone();
 
-        let mut size = pixels_to_points(font_size_to_pixels(self.font_size) * 0.75)
-            * font_scale_from_ctx(sctx);
+        let mut size = font_size_to_pixels(self.pen.font_size) * font_scale_from_ctx(sctx);
         if matches!(self.ruby, Ruby::Over) {
             size /= 2.0;
             result
@@ -303,13 +287,14 @@ impl Segment {
                 .set(OpenTypeTag::FEAT_RUBY, 1);
         }
 
-        // FIXME: This is not correct but *looks* more correct :)
-        *result.make_inline_sizing_mut() = InlineSizing::Stretch;
+        if use_inlines {
+            *result.make_inline_sizing_mut() = InlineSizing::Stretch;
+        }
 
         *result.make_font_size_mut() = I26Dot6::from(size);
 
         let mut shadows = vec![];
-        self.shadow.to_css(sctx, &mut shadows);
+        self.compute_shadows(sctx, &mut shadows);
 
         if !shadows.is_empty() {
             *result.make_text_shadows_mut() = shadows.into();
@@ -319,58 +304,119 @@ impl Segment {
     }
 }
 
-fn segments_to_inline(
-    pass: &mut FrameLayoutPass,
-    content: &mut InlineContentBuilder,
-    event_time: u32,
-    segments: &[Segment],
-) {
-    // What lack of Peekable::inner() and Filter::inner() does to a language...
-    let mut next_idx = 0;
-    let sctx = pass.sctx;
-    let mut root = content.root();
-    let mut it = segments
-        .iter()
-        .filter(|segment| {
-            pass.add_animation_point(event_time + segment.time_offset);
-            segment.time_offset <= pass.t - event_time
-        })
-        .peekable();
+impl VisualLine {
+    fn to_inline_content(
+        &self,
+        pass: &mut FrameLayoutPass,
+        root_inline_style: ComputedStyle,
+    ) -> InlineContent {
+        let mut builder = InlineContentBuilder::new(root_inline_style);
+        // What lack of Peekable::inner() and Filter::inner() does to a language...
+        let mut next_idx = 0;
+        let sctx = pass.sctx;
+        let use_inlines = pass.srv3_use_inlines;
+        let mut root = builder.root();
+        let mut it = self
+            .segments
+            .iter()
+            .filter(|segment| {
+                pass.add_animation_point(self.range.start + segment.time_offset);
+                segment.time_offset <= pass.t - self.range.start
+            })
+            .peekable();
 
-    while let Some(segment) = it.next() {
-        let mut style = segment.compute_style(sctx);
-
-        if next_idx == 0 {
-            *style.make_padding_left_mut() = Length::from_points(style.font_size() / 4);
-        }
-        next_idx += 1;
-        // NOTE: This purposefully ignores whether or not the next segment is
-        //       currently visible as that is what YouTube seems to do.
-        if segments.get(next_idx).is_none() {
-            *style.make_padding_right_mut() = Length::from_points(style.font_size() / 4);
+        struct CurrentBlock {
+            pen: Option<Pen>,
+            style: ComputedStyle,
+            builder: InlineContentBuilder,
         }
 
-        match segment.ruby {
-            Ruby::None => {
-                root.push_span(style).push_text(&segment.text);
+        impl CurrentBlock {
+            fn flush(&mut self, to: &mut InlineSpanBuilder<'_>) {
+                to.push_inline_block(BlockContainer {
+                    style: std::mem::replace(&mut self.style, ComputedStyle::DEFAULT),
+                    content: BlockContainerContent::Inline(self.builder.finish()),
+                });
             }
-            Ruby::Base => {
-                let mut ruby = root.push_ruby(style.create_derived());
-                ruby.push_base(style.clone()).push_text(&segment.text);
-                if let Some(next) = it.peek() {
-                    if let Ruby::Over = next.ruby {
-                        ruby.push_annotation(next.compute_style(sctx))
-                            .push_text(&next.text);
-                        _ = it.next();
+        }
+
+        // Only used if inline-block layout is enabled to coalesce structurally-same-pen
+        // segments into a single inline-block.
+        // This is what YouTube appears to do from my testing.
+        let mut current_block: CurrentBlock = CurrentBlock {
+            pen: None,
+            style: ComputedStyle::DEFAULT,
+            builder: InlineContentBuilder::new(ComputedStyle::DEFAULT),
+        };
+        while let Some(segment) = it.next() {
+            let mut style = segment.compute_style(sctx, use_inlines);
+
+            if next_idx == 0 {
+                *style.make_padding_left_mut() = Length::from_points(style.font_size() / 4);
+            }
+            next_idx += 1;
+            // NOTE: This purposefully ignores whether or not the next segment is
+            //       currently visible as that is what YouTube seems to do.
+            if self.segments.get(next_idx).is_none() {
+                *style.make_padding_right_mut() = Length::from_points(style.font_size() / 4);
+            }
+
+            let mut layout_to_builder =
+                |inner: &mut InlineSpanBuilder<'_>, inner_style: ComputedStyle| match segment.ruby {
+                    Ruby::None => {
+                        inner.push_span(inner_style).push_text(&segment.text);
                     }
-                }
-            }
-            Ruby::Over => {
-                root.push_ruby(style.clone())
-                    .push_annotation(style)
-                    .push_text(&segment.text);
-            }
+                    Ruby::Base => {
+                        let mut ruby = inner.push_ruby(inner_style.create_derived());
+                        ruby.push_base(inner_style.clone()).push_text(&segment.text);
+                        if let Some(next) = it.peek() {
+                            if let Ruby::Over = next.ruby {
+                                let mut annotation_style = next.compute_style(sctx, use_inlines);
+                                // If inline-block layout is enabled then background is handled by the
+                                // containing block.
+                                if !use_inlines {
+                                    *annotation_style.make_background_color_mut() = BGRA8::ZERO;
+                                }
+                                ruby.push_annotation(annotation_style).push_text(&next.text);
+                                _ = it.next();
+                            }
+                        }
+                    }
+                    Ruby::Over => {
+                        inner
+                            .push_ruby(inner_style.create_derived())
+                            .push_annotation(inner_style)
+                            .push_text(&segment.text);
+                    }
+                };
+
+            if use_inlines {
+                layout_to_builder(&mut root, style);
+            } else {
+                let inner_style = style.create_derived();
+                if current_block
+                    .pen
+                    .as_ref()
+                    .is_none_or(|pen| pen != &segment.pen)
+                {
+                    if current_block.pen.is_some() {
+                        current_block.flush(&mut root);
+                    }
+                    current_block.pen = Some(segment.pen);
+                    current_block.builder.set_root_style(style.create_derived());
+                    current_block.style = style;
+                };
+
+                layout_to_builder(&mut current_block.builder.root(), inner_style);
+            };
         }
+
+        if current_block.pen.is_some() {
+            current_block.flush(&mut root);
+        }
+
+        drop(root);
+        builder.finish()
     }
 }
 
@@ -378,29 +424,43 @@ impl Window {
     pub fn layout(
         &self,
         pass: &mut FrameLayoutPass,
-    ) -> Result<Option<(Point2L, layout::inline::InlineContentFragment)>, layout::InlineLayoutError>
+    ) -> Result<Option<(Point2L, layout::block::BlockContainerFragment)>, layout::InlineLayoutError>
     {
-        let mut content = InlineContentBuilder::new();
+        let inner_style = {
+            let mut result = self.segment_style.clone();
+            *result.make_background_color_mut() = BGRA8::ZERO;
+            result
+        };
+        let mut lines = Vec::new();
         for line in &self.lines {
             if pass.add_event_range(line.range.clone()) {
-                if !content.is_empty() {
-                    content.root().push_text("\n");
-                }
-
-                segments_to_inline(pass, &mut content, line.range.start, &line.segments)
+                lines.push(BlockContainer {
+                    style: inner_style.clone(),
+                    content: BlockContainerContent::Inline(
+                        line.to_inline_content(pass, self.segment_style.clone()),
+                    ),
+                });
             }
         }
 
-        if content.is_empty() {
+        if lines.is_empty() {
             return Ok(None);
         }
 
+        let window = BlockContainer {
+            style: inner_style,
+            content: BlockContainerContent::Block(lines),
+        };
+        let partial_window = layout::block::layout_initial(pass.lctx, &window)?;
+
+        let width = partial_window
+            .intrinsic_width()
+            .min(pass.sctx.player_width() * 96 / 100);
         let constraints = LayoutConstraints {
-            size: Vec2L::new(pass.sctx.player_width() * 96 / 100, FixedL::MAX),
+            size: Vec2L::new(width, FixedL::MAX),
         };
 
-        let fragment =
-            layout::inline::layout(pass.lctx, &constraints, &content.finish(), self.alignment.0)?;
+        let fragment = partial_window.layout(pass.lctx, &constraints)?;
 
         let mut pos = Point2L::new(
             (self.x * pass.sctx.player_width().into_f32()).into(),
@@ -408,13 +468,13 @@ impl Window {
         );
 
         let fragment_size = fragment.fbox.size_for_layout();
-        match self.alignment.0 {
+        match self.segment_style.text_align() {
             HorizontalAlignment::Left => (),
             HorizontalAlignment::Center => pos.x -= fragment_size.x / 2,
             HorizontalAlignment::Right => pos.x -= fragment_size.x,
         }
 
-        match self.alignment.1 {
+        match self.vertical_align {
             VerticalAlignment::Top => (),
             VerticalAlignment::Center => pos.y -= fragment_size.y / 2,
             VerticalAlignment::Bottom => pos.y -= fragment_size.y,
@@ -466,14 +526,10 @@ fn convert_segment(
     base_style: &ComputedStyle,
 ) -> Segment {
     Segment {
+        pen: *segment.pen(),
         base_style: pen_to_size_independent_style(segment.pen(), false, base_style.clone()),
-        font_size: segment.pen().font_size,
         time_offset: segment.time_offset,
         text: text.into(),
-        shadow: Srv3TextShadow {
-            kind: segment.pen().edge_type,
-            color: BGRA8::from_argb32(segment.pen().edge_color | 0xFF000000),
-        },
         ruby,
     }
 }
@@ -485,6 +541,21 @@ struct WindowBuilder<'a> {
 }
 
 impl WindowBuilder<'_> {
+    fn create_window(&self, x: f32, y: f32, time: u32, duration: u32, align: Alignment) -> Window {
+        Window {
+            x,
+            y,
+            range: time..time + duration,
+            segment_style: {
+                let mut style = self.base_style.clone();
+                *style.make_text_align_mut() = align.0;
+                style
+            },
+            vertical_align: align.1,
+            lines: Vec::new(),
+        }
+    }
+
     fn extend_lines(&mut self, window: &mut Window, event: &Event) {
         let event_range = event.time..event.time + event.duration;
         let mut current_line = VisualLine {
@@ -523,7 +594,7 @@ impl WindowBuilder<'_> {
                         segment,
                         &segment.text,
                         Ruby::Base,
-                        &self.base_style,
+                        &window.segment_style,
                     ));
                     _ = it.next().unwrap();
                     let next = it.next().unwrap();
@@ -531,7 +602,7 @@ impl WindowBuilder<'_> {
                         next,
                         &next.text,
                         ruby,
-                        &self.base_style,
+                        &window.segment_style,
                     ));
                     _ = it.next().unwrap();
 
@@ -549,7 +620,7 @@ impl WindowBuilder<'_> {
                     segment,
                     &segment.text[last..end],
                     Ruby::None,
-                    &self.base_style,
+                    &window.segment_style,
                 ));
 
                 if end == segment.text.len() {
@@ -569,26 +640,35 @@ impl WindowBuilder<'_> {
     }
 }
 
-pub fn convert(sbr: &Subrandr, document: Document) -> Subtitles {
+pub fn convert(sbr: &Subrandr, document: Document, lang: Option<&LanguageIdentifier>) -> Subtitles {
     let mut result = Subtitles {
         windows: Vec::new(),
     };
     let mut window_builder = WindowBuilder {
         sbr,
-        base_style: pen_to_size_independent_style(&Pen::DEFAULT, true, ComputedStyle::DEFAULT),
+        base_style: {
+            let mut style =
+                pen_to_size_independent_style(&Pen::DEFAULT, true, ComputedStyle::DEFAULT);
+            if let Some(lang) = lang {
+                if LocaleDirectionality::new_common().is_right_to_left(lang) {
+                    *style.make_direction_mut() = Direction::Rtl;
+                }
+            }
+            style
+        },
         logset: LogOnceSet::new(),
     };
 
     let mut wname_to_index = HashMap::new();
     for (name, window) in document.windows() {
         wname_to_index.insert(&**name, result.windows.len());
-        result.windows.push(Window {
-            x: convert_coordinate(window.position().x as f32),
-            y: convert_coordinate(window.position().y as f32),
-            range: window.time..window.time + window.duration,
-            alignment: window.position().point.to_alignment(),
-            lines: Vec::new(),
-        });
+        result.windows.push(window_builder.create_window(
+            convert_coordinate(window.position().x as f32),
+            convert_coordinate(window.position().y as f32),
+            window.time,
+            window.duration,
+            window.position().point.to_alignment(),
+        ));
     }
 
     for event in document.events() {
@@ -610,13 +690,13 @@ pub fn convert(sbr: &Subrandr, document: Document) -> Subtitles {
         {
             window_builder.extend_lines(&mut result.windows[widx], event);
         } else {
-            let mut window = Window {
-                x: convert_coordinate(event.position().x as f32),
-                y: convert_coordinate(event.position().y as f32),
-                range: event.time..event.time + event.duration,
-                alignment: event.position().point.to_alignment(),
-                lines: Vec::new(),
-            };
+            let mut window = window_builder.create_window(
+                convert_coordinate(event.position().x as f32),
+                convert_coordinate(event.position().y as f32),
+                event.time,
+                event.duration,
+                event.position().point.to_alignment(),
+            );
             window_builder.extend_lines(&mut window, event);
             result.windows.push(window);
         }
