@@ -10,6 +10,7 @@ use util::{
 use crate::{
     layout::{
         self,
+        block::BlockContainerFragment,
         inline::{InlineContentBuilder, InlineContentFragment, InlineSpanBuilder, LineBoxFragment},
         FixedL, InlineLayoutError, Point2L, Vec2L,
     },
@@ -123,11 +124,14 @@ impl Event {
                 size: Vec2L::new(sctx.video_width * self.size as f32 / 100, FixedL::MAX),
             },
             &{
-                let mut builder = InlineContentBuilder::new();
+                let mut builder = InlineContentBuilder::new({
+                    let mut style = self.root.base_style.clone();
+                    *style.make_text_align_mut() = self.horizontal_alignment;
+                    style
+                });
                 self.root.append_to(&mut builder.root(), font_size);
                 builder.finish()
             },
-            self.horizontal_alignment,
         )?;
 
         let lines = &mut fragment.lines;
@@ -650,8 +654,8 @@ impl Layouter {
                 continue;
             }
 
-            let (pos, block) = event.layout(pass.sctx, pass.lctx, font_size, &mut output)?;
-            pass.emit_fragment(pos, block);
+            let (pos, inline) = event.layout(pass.sctx, pass.lctx, font_size, &mut output)?;
+            pass.emit_fragment(pos, BlockContainerFragment::from_inline(inline));
         }
 
         Ok(())
