@@ -2,14 +2,17 @@ use std::{path::Path, sync::OnceLock};
 
 use rasterize::{
     color::{to_straight_rgba, Premultiplied, BGRA8},
+    scene::SceneNode,
     Rasterizer,
 };
 use util::rc::{rc_static, Rc};
 
 use crate::{
     display::DisplayPass,
-    layout::{self, inline::InlineContent, LayoutConstraints, LayoutContext, Point2L, Vec2L},
-    style::computed::HorizontalAlignment,
+    layout::{
+        self, block::BlockContainer, inline::InlineContent, LayoutConstraints, LayoutContext,
+        Point2L, Vec2L,
+    },
     text::{Face, FaceInfo, FontDb, GlyphCache},
     Subrandr,
 };
@@ -21,11 +24,10 @@ macro_rules! make_tree {
     { $what: ident $block: tt } => { make_tree!(@build $what [crate::style::ComputedStyle::DEFAULT;]; $block) };
 
     (@build inline [$style: expr;]; { $($content: tt)* }) => {{
-        let mut builder = crate::layout::inline::InlineContentBuilder::new();
+        let mut builder = crate::layout::inline::InlineContentBuilder::new($style.clone());
         {
-            let mut root = builder.root();
             #[allow(unused)] // this is unused if the inline is empty
-            let mut root = root.push_span($style.clone());
+            let mut root = builder.root();
             make_tree!(@build_all inline [$style; inline=root]; $($content)*);
         }
         builder.finish()
@@ -39,6 +41,10 @@ macro_rules! make_tree {
     (@map_child inline->ruby $value: expr) => {
         $value
     };
+    (@map_child inline->block $value: expr) => {
+        $value
+    };
+    (@build_all_result_ty inline) => { () };
 
     (@build text [$style: expr; inline=$builder: ident]; $value: literal) => {
         $builder.push_text($value)
@@ -52,6 +58,14 @@ macro_rules! make_tree {
         let mut builder = $builder.push_ruby($style.clone());
         make_tree!(@build_all ruby [$style; inline=builder]; $($content)*);
     }};
+    (@build block [$style: expr; inline=$builder: ident]; $content_block: tt) => {{
+        $builder.push_inline_block(
+            crate::layout::block::BlockContainer {
+                style: $style.clone(),
+                content: make_tree!(@build_block_content [$style;] $content_block)
+            }
+        );
+    }};
     (@map_child ruby->base $value: expr) => {
         $value
     };
@@ -66,6 +80,28 @@ macro_rules! make_tree {
         let mut builder = $builder.push_annotation($style.clone());
         make_tree!(@build_all inline [$style; inline=builder]; $($content)*);
     }};
+    (@build_all_result_ty ruby) => { () };
+
+    (@build block [$style: expr;]; $content_block: tt) => {{
+        crate::layout::block::BlockContainer {
+            style: $style.clone(),
+            content: make_tree!(@build_block_content [$style;] $content_block)
+        }
+    }};
+    (@build_block_content [$style: expr;] { inline $(.$class: ident)* { $($content: tt)* } }) => {
+        crate::layout::block::BlockContainerContent::Inline(
+            make_tree!(@build inline $(.$class)*  [$style;]; { $($content)* })
+        )
+    };
+    (@build_block_content [$style: expr;] { $($content: tt)* }) => {
+        crate::layout::block::BlockContainerContent::Block(
+            make_tree!(@build_all block [$style;]; $($content)*).into()
+        )
+    };
+    (@map_child block->block $value: expr) => {
+        $value
+    };
+    (@build_all_result_ty block) => { crate::layout::block::BlockContainer };
 
     (@build $what: ident [$parent_style: expr; $($context_rest: tt)*]; [$($class: ident)*] { $($block_content: tt)* }) => {{
         let style = make_tree!(@apply_style $parent_style; $($class)*);
@@ -118,7 +154,7 @@ macro_rules! make_tree {
             ),
         ]; $($rest)*)
     };
-    (@build_all_rec $mapper: ident $context: tt $result: tt;) => { $result as [(); _] };
+    (@build_all_rec $mapper: ident $context: tt $result: tt;) => { $result as [make_tree!(@build_all_result_ty $mapper); _] };
     (@map_child $from: ident->$to: ident $value: expr) => {
         compile_error!(concat!(
             stringify!($from),
@@ -200,13 +236,11 @@ test_define_style! {
     pub .noto_color_emoji { font_family: rc_static!([NOTO_COLOR_EMOJI.family()])}
 }
 
-pub fn check_inline(
+fn check_fn(
     name: &str,
-    pos: Point2L,
     viewport_size: Vec2L,
-    align: HorizontalAlignment,
-    inline: InlineContent,
     dpi: u32,
+    fun: impl FnOnce(&mut LayoutContext, &LayoutConstraints, &mut Vec<SceneNode>),
 ) {
     let project_dir = test_util::project_dir();
     let tests_dir = project_dir.join("tests/");
@@ -224,7 +258,8 @@ pub fn check_inline(
     let width = viewport_size.x.ceil_to_inner() as u32;
     let height = viewport_size.y.ceil_to_inner() as u32;
     let mut pixels = {
-        let fragment = layout::inline::layout(
+        let mut scene = Vec::new();
+        fun(
             &mut LayoutContext {
                 dpi,
                 fonts: &mut fonts,
@@ -232,17 +267,11 @@ pub fn check_inline(
             &LayoutConstraints {
                 size: viewport_size,
             },
-            &inline,
-            align,
-        )
-        .expect("Inline layout failed");
+            &mut scene,
+        );
 
         let mut pixels = vec![Premultiplied(BGRA8::ZERO); width as usize * height as usize];
         let glyph_cache = GlyphCache::new();
-        let mut scene = Vec::new();
-
-        DisplayPass { output: &mut scene }.display_inline_content_fragment(pos, &fragment);
-
         let mut rasterizer = rasterize::sw::Rasterizer::new();
         let mut render_target =
             rasterize::sw::RenderTarget::new(&mut pixels, width, height, width).into();
@@ -264,10 +293,38 @@ pub fn check_inline(
     );
 }
 
+pub fn check_inline(
+    name: &str,
+    pos: Point2L,
+    viewport_size: Vec2L,
+    dpi: u32,
+    inline: InlineContent,
+) {
+    check_fn(name, viewport_size, dpi, |lctx, constraints, output| {
+        let fragment =
+            layout::inline::layout(lctx, constraints, &inline).expect("Inline layout failed");
+
+        DisplayPass::new(output, dpi).display_inline_content_fragment(pos, &fragment);
+    })
+}
+
+pub fn check_block(
+    name: &str,
+    pos: Point2L,
+    viewport_size: Vec2L,
+    dpi: u32,
+    block: BlockContainer,
+) {
+    check_fn(name, viewport_size, dpi, |lctx, constraints, output| {
+        let fragment = layout::block::layout(lctx, constraints, &block).expect("Layout failed");
+
+        DisplayPass::new(output, dpi).display_block_container_fragment(pos, &fragment);
+    })
+}
+
 macro_rules! check_one {
     (
         name = $name: expr,
-        $(align = $align: ident,)?
         $(dpi = $dpi: literal,)?
         $(pos = $p: tt,)?
         size = ($sx: expr, $sy: expr),
@@ -279,7 +336,6 @@ macro_rules! check_one {
         let submodule_start = module_path!().find(TEST_ROOT_MODULE).unwrap() + TEST_ROOT_MODULE.len();
         let prefix = module_path!()[submodule_start..].replace("::", "_");
         let name = format!("{prefix}_{}", $name);
-        let align = check_one!(@align $($align)?);
         let dpi = check_one!(@dpi $($dpi)?);
         let pos = check_one!(@pos $($p)?);
         let size = $crate::layout::Vec2L::new(
@@ -290,10 +346,10 @@ macro_rules! check_one {
         use $crate::layout_tests::common::make_tree;
         let tree = make_tree!($what $(.$class)+ { $($block_content)* }) ;
 
-        check_inline(&name, pos, size, align, tree, dpi);
+        check_one!(@check $what (&name, pos, size, dpi, tree));
     }};
-    (@align $name: ident) => { $crate::style::computed::HorizontalAlignment::$name };
-    (@align) => { check_one!(@align Left) };
+    (@check inline $args: tt) => { check_inline $args; };
+    (@check block $args: tt) => { check_block $args; };
     (@dpi $value: literal) => { $value };
     (@dpi) => { 72 };
     (@pos) => { $crate::layout::Point2L::ZERO };
