@@ -1,10 +1,14 @@
 pub mod common;
 
+mod block;
 mod font_feature_settings;
 mod hello_world;
+mod inline_block;
 mod inline_sizing;
 mod input_edge_cases;
 mod line_break_anywhere;
 mod padding;
 mod ruby;
+mod text_decoration;
 mod text_shadow;
+mod whitespace;
